/*----------------------------------------------------------------------------
;
; Windows keystroke event reader -- Paul Houle (paulhoule.com) 12/31/2019
;
; Echoes keyboard keystroke event info to console in hex as a key is
; pressed (and optionally released).
;
; The info is echoed as a hex number whose bit fields are as follows:
;
;	Key Down			(25)
;	Enhanced Key		(24)
;	Caps Lock On		(23)
;	Scroll Lock On		(22)
;	Num Lock On			(21)
;	Shift				(20)
;	Left Ctrl			(19)
;	Right Ctrl			(18)
;	Left Alt			(17)
;	Right Alt			(16)
;	Translated Ascii	(15..8)
;	Virtual Key Code	 (7..0)
;
; By default only key downstrokes are echoed until an <Esc> is entered.
;
; Options that can be added to the command line:
;
;	phkev /				include key release events
;	phkev *				do not display event, return it as program exitcode
;
; The * mode will return immediately upon the first event (which is not
; echoed in hex).  An example of its use:
;
;	set errorlevel=& phkev *
;	echo %errorlevel%
;	33561371
;
; The usefulness of * might be, via some .bat set /a arithmetic, to isolate
; the scancode and check for user input of a special character (like a
; function or arrow key).
;
; The .exe was compiled with Microsoft cl ver 13.10.3077 (7.1) via:
;
;	cl /Oxb0 phkev.c /link /merge:.data=.text /merge:.rdata=.text
;		 /ignore:4078 /entry:main /filealign:512
;
; The program can be compiled with a more recent compiler, or 64-bit, but
; the resulting .exe will be larger.
;
;---------------------------------------------------------------------------*/
#include <windows.h>
#pragma comment(lib, "kernel32")

// Returns options based on special chars in command line.
#define OPT_RELEASE	1					// don't ignore key release events
#define OPT_UNUSED	2
#define OPT_RETKEV	4					// no display, return kev as exit code
char GetOpts(void) {
	char ch, *pch, opts = 0;

	if (pch = GetCommandLineA()) {
		while (ch = *pch++) {
			DWORD optch = '*'<<16 | '?'<<8 | '/';
			char optfl = 1;
			do {
				if (ch == (char) optch) opts |= optfl;
			} while (optfl <<= 1, optch >>= 8);
		}
	}
	return opts;
}

/*
	Return single keyboard event packed to int, or -1 if error.
*/
int GetKev() {
	INPUT_RECORD InputRecord;
	DWORD NumRecord;
	int kev = -1;
	do {
		if (!ReadConsoleInputA(GetStdHandle(STD_INPUT_HANDLE),
			&InputRecord, 1, &NumRecord) || !NumRecord) break;
		if (InputRecord.EventType == KEY_EVENT) {
			kev = InputRecord.Event.KeyEvent.bKeyDown;
			kev = kev << 9 | InputRecord.Event.KeyEvent.dwControlKeyState;
			kev = kev << 8 | InputRecord.Event.KeyEvent.uChar.AsciiChar;
			kev = kev << 8 | InputRecord.Event.KeyEvent.wVirtualKeyCode;
		}
	} while (kev == -1);
	return kev;
}

/*
	Displays compacted keyboard event in hex, or banner if kev == -1.
	Returns non-zero for success.
*/
char banner[] =	"Displays keystroke events.  <Esc> to exit."
	"  paulhoule.com/phkev  12/31/2019\r\n";
int __fastcall DisplayKev(int kev) {
	DWORD nxf;
	char ch, txb[8], *pch;
	int i;

	pch = banner, i = sizeof(banner) - 1;
	if (kev != -1) {
		pch = txb + (i = 7);
		do {
			ch = ((char) kev & 0xf) + '0';
			kev >>= 4;
			if (ch > '9') ch += 'a' - ('9' + 1);
			*pch-- = ch;
		} while (--i);
		*pch = ' ';
		i += 8;
	}
	return WriteFile(GetStdHandle(STD_OUTPUT_HANDLE), pch, i, &nxf, NULL);
}

int main(void) {
	int kev;
	char opts = GetOpts();

	if (!(opts & OPT_RETKEV)) DisplayKev(-1);
	do {
		do {
			kev = GetKev();
			if (kev == -1) return -1;
		} while (!(kev & 1 << 25) && !(opts & OPT_RELEASE));
		if (opts & OPT_RETKEV) return kev;

		if (!DisplayKev(kev)) return -1;
	} while ((char) kev != 0x1b);
	return 0;
}
